from datetime import timedelta
from typing import List, Dict, Tuple, Any

from pytest import fixture, mark

from resotocore.db.runningtaskdb import RunningTaskData, RunningTaskDb, RunningTaskStepInfo
from resotocore.ids import TaskId, SubscriberId, TaskDescriptorId
from resotocore.message_bus import ActionDone
from resotocore.task.model import Subscriber
from resotocore.task.task_description import RunningTask, Workflow
from resotocore.util import utc, utc_str

now = utc()


@fixture
def instances() -> List[RunningTaskData]:
    messages = [ActionDone(str(a), TaskId("test"), "bla", SubscriberId("sf")) for a in range(0, 10)]
    state_data = {"test": 1}
    return [
        RunningTaskData(
            TaskId(f"task_{a}"),
            TaskDescriptorId("task_123"),
            "task_123",
            Workflow.__name__,
            messages,
            "start",
            state_data,
            [RunningTaskStepInfo(f"step_{a}", False, now, now) for a in range(0, 3)],
            task_started_at=now,
            task_duration=timedelta(seconds=10),
            done=a > 5,
            has_info=a > 6,
            has_error=a > 7,
        )
        for a in range(0, 10)
    ]


@mark.asyncio
async def test_load_running(running_task_db: RunningTaskDb, instances: List[RunningTaskData]) -> None:
    await running_task_db.update_many(instances)
    not_done = list(filter(lambda x: not x.done, instances))
    assert not_done.sort() == [sub async for sub in running_task_db.all_running()].sort()


@mark.asyncio
async def test_last(running_task_db: RunningTaskDb, instances: List[RunningTaskData]) -> None:
    await running_task_db.update_many(instances)
    running_tasks = list(filter(lambda x: x.done, instances))
    running_tasks.sort(key=lambda x: x.task_started_at, reverse=True)
    done_task = next(iter(running_tasks), None)
    assert done_task
    last_done = await running_task_db.last(descriptor_id=done_task.task_descriptor_id)
    assert last_done
    assert done_task.id == last_done.id

    assert await running_task_db.last()


@mark.asyncio
async def test_load(running_task_db: RunningTaskDb, instances: List[RunningTaskData]) -> None:
    await running_task_db.update_many(instances)
    loaded = [sub async for sub in running_task_db.all()]
    assert instances.sort() == loaded.sort()


@mark.asyncio
async def test_filtered(running_task_db: RunningTaskDb, instances: List[RunningTaskData]) -> None:
    await running_task_db.update_many(instances)

    async def filtered_list(**kwargs: Any) -> List[TaskId]:
        async with await running_task_db.filtered(**kwargs) as crsr:
            return [elem.id async for elem in crsr]

    assert len(await filtered_list()) == len(instances)
    assert len(await filtered_list(limit=1)) == 1
    assert len(await filtered_list(descriptor_id="task_123")) == len(instances)
    assert await filtered_list(task_id=TaskId("task_1")) == [TaskId("task_1")]
    assert len(await filtered_list(started_after=now + timedelta(minutes=1))) == 0
    assert len(await filtered_list(started_after=now - timedelta(minutes=1))) == 10
    assert len(await filtered_list(started_before=now + timedelta(minutes=1))) == 10
    assert len(await filtered_list(started_before=now - timedelta(minutes=1))) == 0
    assert len(await filtered_list(with_info=True)) == 3
    assert len(await filtered_list(with_info=False)) == 7
    assert len(await filtered_list(with_error=True)) == 2
    assert len(await filtered_list(with_error=False)) == 8


@mark.asyncio
async def test_aggregated(running_task_db: RunningTaskDb, instances: List[RunningTaskData]) -> None:
    await running_task_db.update_many(instances)
    res = await running_task_db.aggregated_history()
    assert res == {
        "task_123": {"count": 10, "last_run": utc_str(now), "runs_with_errors": 2, "average_duration": "10s"}
    }


@mark.asyncio
async def test_update(running_task_db: RunningTaskDb, instances: List[RunningTaskData]) -> None:
    # multiple updates should work as expected
    await running_task_db.update_many(instances)
    await running_task_db.update_many(instances)
    await running_task_db.update_many(instances)
    loaded = [sub async for sub in running_task_db.all()]
    assert instances.sort() == loaded.sort()


@mark.asyncio
async def test_delete(running_task_db: RunningTaskDb, instances: List[RunningTaskData]) -> None:
    await running_task_db.update_many(instances)
    remaining = list(instances)
    for _ in instances:
        sub = remaining.pop()
        await running_task_db.delete_value(sub)
        loaded = [sub async for sub in running_task_db.all()]
        assert remaining.sort() == loaded.sort()
    assert len([sub async for sub in running_task_db.all()]) == 0


@mark.asyncio
async def test_update_state(
    running_task_db: RunningTaskDb,
    workflow_instance: Tuple[RunningTask, Subscriber, Subscriber, Dict[str, List[Subscriber]]],
) -> None:
    wi, _, _, _ = workflow_instance
    task_id = TaskId("test")
    subscriber_id = SubscriberId("sf")
    first = ActionDone("start_collect", task_id, "bla", subscriber_id)
    second = ActionDone("collect", task_id, "bla", subscriber_id)
    third = ActionDone("collect_done", task_id, "bla", subscriber_id)

    async def assert_state(current: str, message_count: int) -> RunningTaskData:
        state: RunningTaskData = await running_task_db.get(wi.id)  # type: ignore
        assert state.current_state_name == current
        assert len(state.received_messages) == message_count
        return state

    await running_task_db.insert(wi)
    await assert_state(wi.current_state.name, 6)

    wi.machine.set_state("start")
    await running_task_db.update_state(wi, first)
    await assert_state("start", 7)

    wi.machine.set_state("collect")
    await running_task_db.update_state(wi, second)
    await assert_state("collect", 8)

    wi.machine.set_state("done")
    await running_task_db.update_state(wi, third)
    last = await assert_state("done", 9)

    assert last.received_messages[-3:] == [first, second, third]



from typing import AsyncIterator, List, cast
from resotocore.graph_manager.graph_manager import GraphManager, _compress_timestamps
from resotocore.ids import GraphName
from resotocore.db.db_access import DbAccess
from resotocore.db.model import GraphUpdate
from resotocore.model.model import Model
from resotocore.types import Json
from resotocore.config.core_config_handler import CoreConfigHandler
from resotocore.core_config import SnapshotsScheduleConfig, SnapshotSchedule, ResotoCoreSnapshotsConfigId
from resotocore.task import TaskHandler
import pytest
from tests.resotocore.db.graphdb_test import create_multi_collector_graph
import re
from resotolib.utils import utc
from datetime import timedelta


@pytest.mark.asyncio
async def test_graph_manager(
    foo_model: Model,
    db_access: DbAccess,
    core_config_handler: CoreConfigHandler,
    task_handler: TaskHandler,
) -> None:
    # test setup
    graph_name = GraphName("test_graph")
    graph_db = await db_access.create_graph(graph_name, validate_name=False)
    await graph_db.wipe()
    await db_access.delete_graph(GraphName("test_graph_copy"))

    # test timestamp compression:
    assert _compress_timestamps("test_graph") == "test_graph"
    assert _compress_timestamps("foo123") == "foo123"
    assert _compress_timestamps("foo_2023-05-24T14:06:50+0000") == "foo_20230524T140650Z"
    assert _compress_timestamps("2023-05-24T14:06:50+0000_foo") == "20230524T140650Z_foo"
    assert (
        _compress_timestamps("foo_2023-05-24T14:06:50+0000_bar_2023-05-24T14:06:50+0000")
        == "foo_20230524T140650Z_bar_20230524T140650Z"
    )
    assert (
        _compress_timestamps("foo-bar-graph-label-2023-05-24T14:06:50+0000") == "foo-bar-graph-label-20230524T140650Z"
    )

    # populate some data in the graphes
    nodes, info = await graph_db.merge_graph(create_multi_collector_graph(), foo_model)
    assert info == GraphUpdate(110, 1, 0, 218, 0, 0)
    assert len(nodes) == 8

    graph_manager = GraphManager(db_access, SnapshotsScheduleConfig(snapshots={}), core_config_handler, task_handler)
    await graph_manager.start()

    # list
    assert GraphName("test_graph") in await graph_manager.list(".*")

    # copy
    await graph_manager.copy(GraphName("test_graph"), GraphName("test_graph_copy"), False, False)
    assert set(await graph_manager.list(".*")).issuperset(["test_graph", "test_graph_copy"])

    # snapshot
    now = utc()
    await graph_manager.snapshot(GraphName("test_graph"), "label", timestamp=now)
    graphs = await graph_manager.list(".*")
    for name in ["test_graph", "test_graph_copy", "snapshot-test_graph-label-.*"]:
        for graph in graphs:
            if re.match(name, graph):
                break
        else:
            raise AssertionError(f"Could not find graph with name {name} in {graphs}")

    # snapshots at specific time
    now = now + timedelta(seconds=42)
    await graph_manager.snapshot(GraphName("test_graph"), "label", timestamp=now)
    somewhere_in_the_past = now
    now = now + timedelta(seconds=42)
    await graph_manager.snapshot(GraphName("test_graph"), "label", timestamp=now)
    now = now + timedelta(seconds=42)
    await graph_manager.snapshot(GraphName("test_graph"), "label", timestamp=now)
    found = await graph_manager.snapshot_at(time=somewhere_in_the_past, graph_name=GraphName("test_graph"))
    assert found is not None

    # test snapshot cleanup
    await graph_manager._clean_outdated_snapshots(
        SnapshotsScheduleConfig(snapshots={"label": SnapshotSchedule("0 1 2 3 4", 0)})
    )
    graphs = await graph_manager.list(".*")
    for graph in graphs:
        if re.match("snapshot-test_graph-label-.*", graph):
            raise AssertionError(f"Found outdated snapshot {graph} in {graphs}")

    # no snapshots after the cleanup
    assert await graph_manager.snapshot_at(time=somewhere_in_the_past, graph_name=GraphName("test_graph")) is None

    # delete
    await graph_manager.delete(GraphName("test_graph_copy"))
    assert GraphName("test_graph_copy") not in set(await graph_manager.list(".*"))

    # periodic snapshot callback
    for job in await task_handler.list_jobs():
        await task_handler.delete_job(job.id, force=True)
    jobs = await task_handler.list_jobs()
    # let's schedule something
    await graph_manager._on_config_updated(ResotoCoreSnapshotsConfigId)
    jobs = await task_handler.list_jobs()
    assert len(jobs) == 5
    assert jobs[0].name == "resoto:snapshots:hourly"

    # test export and import
    dump = []
    test_graph_model = await db_access.get_graph_model_db(GraphName("test_graph"))
    await test_graph_model.update_many(list(foo_model.kinds.values()))
    async for string in graph_manager.export_graph(GraphName("test_graph")):
        dump.append(string)

    async def dump_iter() -> AsyncIterator[str]:
        for string in dump:
            yield string

    await graph_manager.import_graph(GraphName("test_graph_import"), dump_iter(), True)
    # check vertices count
    original_vertices_count = await db_access.db.count("test_graph")
    imported_vertices_count = await db_access.db.count("test_graph_import")
    assert original_vertices_count == imported_vertices_count

    async def collect_docs(graph_name: str) -> List[Json]:
        vertices = []
        async with await db_access.db.aql_cursor(f"for v in {graph_name} return v") as cursor:
            async for vertex in cursor:
                vertex = cast(Json, vertex)
                if vertex.get("_key") == "root":
                    continue
                # remove the fields that are going to be different after the import
                vertex.pop("_id")
                vertex.pop("_rev")
                vertex.pop("hash", None)
                vertex.pop("_from", None)
                vertex.pop("_to", None)
                vertices.append(vertex)

        return sorted(vertices, key=lambda doc: doc["_key"])

    # check vertices content after import
    original_vertices = await collect_docs("test_graph")
    imported_vertices = await collect_docs("test_graph_import")
    assert original_vertices == imported_vertices

    # check default edges after import
    assert await db_access.db.count("test_graph_default") == await db_access.db.count("test_graph_import_default")
    original_default_edges = await collect_docs("test_graph_default")
    imported_default_edges = await collect_docs("test_graph_import_default")
    assert original_default_edges == imported_default_edges

    # check delete edges after import
    assert await db_access.db.count("test_graph_delete") == await db_access.db.count("test_graph_import_delete")
    orignal_delete_edges = await collect_docs("test_graph_delete")
    imported_delete_edges = await collect_docs("test_graph_import_delete")
    assert orignal_delete_edges == imported_delete_edges

    # check model after import
    assert await db_access.db.count("test_graph_model") == await db_access.db.count("test_graph_import_model")
    assert await collect_docs("test_graph_model") == await collect_docs("test_graph_import_model")

    # cleanup
    for graph in await graph_manager.list(".*"):
        await db_access.delete_graph(graph)
        await db_access.delete_graph_model(graph)



from attrs import frozen
from resotocore.infra_apps.manifest import AppManifest
from resotocore.ids import InfraAppName
from resotocore.db.async_arangodb import AsyncArangoDB
from resotocore.db.entitydb import EntityDb, ArangoEntityDb


@frozen
class InfraAppPackage:
    manifest: AppManifest
    source_url: str


PackageEntityDb = EntityDb[InfraAppName, InfraAppPackage]


def app_package_entity_db(db: AsyncArangoDB, collection: str) -> ArangoEntityDb[InfraAppName, InfraAppPackage]:
    return ArangoEntityDb(db, collection, InfraAppPackage, lambda k: InfraAppName(k.manifest.name))



import logging
from asyncio import sleep
from functools import lru_cache
from socket import gethostname
from typing import Callable, Awaitable

from aiohttp import ClientConnectionError, ClientResponse
from aiohttp.web import HTTPNotFound, Request, StreamResponse
from aiohttp.web_exceptions import HTTPBadGateway
from multidict import CIMultiDict

from resotocore.web import api  # pylint: disable=unused-import # prevent circular import
from resotocore.web.directives import enable_compression

log = logging.getLogger(__name__)


@lru_cache(1)
def hostname() -> str:
    try:
        return gethostname()
    except Exception:
        return "localhost"


# Proxy request to configured tsdb endpoint
def tsdb(api_handler: "api.Api") -> Callable[[Request], Awaitable[StreamResponse]]:
    def drop_request_specific_headers(headers: CIMultiDict[str]) -> None:
        # since we stream the content (chunked), we are not allowed to set the content length.
        headers.popall("Content-Length", None)
        headers.popall("Content-Encoding", None)
        headers.popall("Connection", None)
        headers.popall("Transfer-Encoding", None)
        headers.popall("Accept-Encoding", None)
        headers.popall("Host", None)

    def should_be_retried(response: ClientResponse) -> bool:
        # We see valid requests failing in k8s with 400.
        # Prometheus will send a content-length header in such a case.
        # Retry the request, if it is missing, since it is not coming from prometheus.
        if response.status == 400 and response.content_length in (None, 0):
            return True
        elif response.status >= 500:
            return True
        else:
            return False

    async def proxy_request(request: Request) -> StreamResponse:
        if api_handler.deps.config.api.tsdb_proxy_url:
            in_headers = request.headers.copy()
            drop_request_specific_headers(in_headers)
            url = f'{api_handler.deps.config.api.tsdb_proxy_url}/{request.match_info["tail"]}'
            max_retries = 5

            async def do_request(attempts_left: int) -> StreamResponse:
                async with api_handler.session.request(
                    request.method,
                    url,
                    params=request.query,
                    headers=in_headers,
                    compress="deflate",
                    data=request.content,
                    ssl=api_handler.deps.cert_handler.client_context,
                ) as cr:
                    try:
                        # in case of error: do we need to retry?
                        if cr.status >= 400 and attempts_left > 0 and should_be_retried(cr):
                            req_header = ", ".join(f"{k}={v}" for k, v in in_headers.items())
                            req_params = ", ".join(f"{k}={v}" for k, v in request.query.items())
                            req_body = await request.content.read()
                            resp_header = ", ".join(f"{k}={v}" for k, v in cr.headers.items())
                            resp_body = await cr.text()
                            log.warning(
                                f"tsdb server returned an error: url:{url}. "
                                f"Request(headers:{req_header}, params:{req_params}, body:{str(req_body)}) "
                                f"Response(headers:{resp_header}, status:{cr.status}, body:{resp_body})"
                            )
                            cr.close()  # close the connection explicitly, might be pooled otherwise
                            await sleep(2 ** (max_retries - attempts_left) * 0.1)  # exponential backoff
                            return await do_request(attempts_left - 1)
                        else:
                            log.info(f"Proxy tsdb request to: {url} resulted in status={cr.status}")
                            headers = cr.headers.copy()
                            drop_request_specific_headers(headers)
                            via = f"{request.version.major}.{request.version.minor} {hostname()}"
                            headers["Via"] = via
                            headers["ViaResoto"] = via  # the via header might be set by other instances ase well
                            response = StreamResponse(status=cr.status, reason=cr.reason, headers=headers)
                            enable_compression(request, response)
                            await response.prepare(request)
                            async for data in cr.content.iter_chunked(1024 * 1024):
                                await response.write(data)
                            await response.write_eof()
                            return response
                    except Exception:
                        cr.close()  # close connection on any error
                        raise

            try:
                return await do_request(attempts_left=max_retries)
            except ClientConnectionError as e:
                log.warning(f"Proxy tsdb request to: {url} is not reachable {e}")
                raise HTTPBadGateway(text="tsdb server is not reachable") from e
        else:
            raise HTTPNotFound(text="No tsdb defined. Adjust resoto.core configuration.")

    return proxy_request



from ipaddress import IPv4Network

from arango.typings import Json

from resotocore.db.model import QueryModel
from resotocore.query.model import FunctionTerm


def has_desired_change(cursor: str, fn: FunctionTerm) -> str:
    return (
        f"{cursor}.desired.{fn.property_path}!=null && "
        + f"{cursor}.reported.{fn.property_path}!={cursor}.desired.{fn.property_path}"
    )


def in_subnet(cursor: str, bind_vars: Json, fn: FunctionTerm, model: QueryModel) -> str:
    """
    Assumptions and requirements:
    - this fn only works for IPv4 addresses
    - ip addresses are stored as 4 octet with digit string in the reported section
    - one argument is given which defines the ip/mask
    :param cursor: the cursor to read from
    :param bind_vars: the bind_vars to send to arango
    :param fn: the function definition
    :param model: the related query model.
    :return: the AQL filter statement
    """
    if len(fn.args) != 1:
        raise AttributeError("Function in_subnet expects exactly one argument. Example: 1.2.3.4/24")
    network = IPv4Network(fn.args[0], strict=False)
    mask = int(network.netmask)
    expected = int(network.network_address) & mask
    length = str(len(bind_vars))
    bind_vars[length] = expected
    return f"BIT_AND(IPV4_TO_NUMBER({cursor}.{fn.property_path}), {mask}) == @{length}"


def has_key(cursor: str, bind_vars: Json, fn: FunctionTerm, model: QueryModel) -> str:
    assert (
        len(fn.args) == 1
    ), "has_key(path.to.property, name_of_prop) or has_key(path.to.property, [name_of_prop_a, name_of_prop_b])"
    args = [fn.args[0]] if isinstance(fn.args[0], str) else fn.args[0]
    for arg in args:
        assert isinstance(arg, str), f"has_key: argument must be string, but got: {arg}"
    prop = f"fn{len(bind_vars)}"
    if len(args) == 0:
        return "true"
    elif len(args) == 1:
        bind_vars[prop] = fn.args[0]
        return f"HAS({cursor}.{fn.property_path}, @{prop})"
    else:
        bind_vars[prop] = args
        return f"@{prop} ALL IN ATTRIBUTES({cursor}.{fn.property_path}, true)"


def as_arangodb_function(cursor: str, bind_vars: Json, fn: FunctionTerm, model: QueryModel) -> str:
    if fn.fn == "has_key":
        return has_key(cursor, bind_vars, fn, model)
    if fn.fn == "in_subnet":
        return in_subnet(cursor, bind_vars, fn, model)
    if fn.fn == "has_desired_change":
        return has_desired_change(cursor, fn)
    else:
        raise AttributeError(f"Function {fn} does not exist!")



import asyncio
from datetime import timedelta, datetime, timezone
from typing import Any, Type, List

from deepdiff import DeepDiff
from pytest import mark

from resotocore.ids import SubscriberId
from resotocore.ids import TaskId
from resotocore.message_bus import (
    MessageBus,
    Message,
    Event,
    Action,
    ActionDone,
    ActionError,
    ActionInfo,
    ActionProgress,
)
from resotocore.model.typed_model import to_js, from_js
from resotocore.util import AnyT, utc, first
from resotolib.core.progress import ProgressDone, Progress


async def wait_for_message(
    all_events: List[Message], message_type: str, t: Type[AnyT], timeout: timedelta = timedelta(seconds=1)
) -> AnyT:
    stop_at = utc() + timeout

    async def find() -> AnyT:
        result = first(lambda m: isinstance(m, t) and m.message_type == message_type, all_events)  # type: ignore
        if result:
            return result  # type: ignore
        elif utc() > stop_at:
            raise TimeoutError()
        else:
            await asyncio.sleep(0.1)
            return await find()

    return await find()


@mark.asyncio
async def test_handler(message_bus: MessageBus) -> None:
    foos: List[Message] = []
    blas: List[Message] = []

    async def emit() -> None:
        await message_bus.emit(Event("foo"))
        await message_bus.emit(Event("foo"))
        await message_bus.emit(Event("bla"))
        await message_bus.emit(Event("bar"))

    async def wait_for(name: str, list: List[Message]) -> None:
        async with message_bus.subscribe(SubscriberId("test"), [name]) as events:
            while True:
                list.append(await events.get())

    foo_t = asyncio.create_task(wait_for("foo", foos))
    bla_t = asyncio.create_task(wait_for("bla", blas))

    await asyncio.sleep(0.1)
    await emit()
    await asyncio.sleep(0.1)
    assert len(foos) == 2
    assert len(blas) == 1
    foo_t.cancel()
    await emit()
    await asyncio.sleep(0.1)
    assert len(foos) == 2
    assert len(blas) == 2
    bla_t.cancel()


def test_message_serialization() -> None:
    task_id = TaskId("123")
    subsctiber_id = SubscriberId("sub")
    now = datetime(2022, 10, 23, 12, 0, 0, 0, timezone.utc)
    roundtrip(Event("test", {"a": "b", "c": 1, "d": "bla"}))
    roundtrip(Action("test", task_id, "step_name"))
    roundtrip(Action("test", task_id, "step_name", {"test": 1}))
    roundtrip(ActionDone("test", task_id, "step_name", subsctiber_id))
    roundtrip(ActionDone("test", task_id, "step_name", subsctiber_id, {"test": 1}))
    roundtrip(ActionError("test", task_id, "step_name", subsctiber_id, "oops"))
    roundtrip(ActionError("test", task_id, "step_name", subsctiber_id, "oops", {"test": 23}))
    roundtrip(ActionInfo("test", task_id, "step_name", subsctiber_id, "error", "Error message"))
    roundtrip(ActionProgress("test", task_id, "step_name", subsctiber_id, ProgressDone("region", 1, 2), now))
    nested = Progress.from_progresses("account1", [ProgressDone("region", 1, 2)])
    pg = ActionProgress("test", task_id, "step_name", subsctiber_id, nested, now)
    assert to_js(pg) == to_js(from_js(to_js(pg), ActionProgress))


def roundtrip(obj: Any) -> None:
    js = to_js(obj)
    again = from_js(js, type(obj))
    assert DeepDiff(obj, again) == {}, f"Json: {js} serialized as {again}"



from datetime import timedelta
from typing import Any, List, Dict, Tuple, Callable

from attr import evolve
from deepdiff import DeepDiff
from frozendict import frozendict

from resotocore.ids import SubscriberId, TaskDescriptorId
from resotocore.message_bus import Action, ActionDone, ActionError, Event, ActionProgress, ActionInfo
from resotocore.model.typed_model import from_js, to_js
from resotocore.task.task_dependencies import TaskDependencies
from resotocore.task.model import Subscriber, Subscription
from resotocore.task.task_description import (
    Workflow,
    Step,
    RunningTask,
    StepErrorBehaviour,
    PerformAction,
    WaitForEvent,
    EmitEvent,
    ExecuteCommand,
    EventTrigger,
    Job,
    TimeTrigger,
    SendMessage,
    ExecuteOnCLI,
)
from resotolib.core.progress import ProgressDone
from resotolib.utils import utc


def test_eq() -> None:
    s1 = Step("a", PerformAction("a"), timedelta())
    s2 = Step("a", WaitForEvent("a", {"foo": "bla"}), timedelta())
    s3 = Step("a", EmitEvent(Event("a", {"a": "b"})), timedelta())
    s4 = Step("a", ExecuteCommand("echo hello"), timedelta())
    assert s1 == Step("a", PerformAction("a"), timedelta())
    assert s2 == Step("a", WaitForEvent("a", {"foo": "bla"}), timedelta())
    assert s3 == Step("a", EmitEvent(Event("a", {"a": "b"})), timedelta())
    assert s4 == Step("a", ExecuteCommand("echo hello"), timedelta())
    trigger = [EventTrigger("start me up")]
    assert Workflow(TaskDescriptorId("a"), "a", [s1, s2, s3, s4], trigger) == Workflow(
        TaskDescriptorId("a"), "a", [s1, s2, s3, s4], trigger
    )


def test_ack_for(workflow_instance: Tuple[RunningTask, Subscriber, Subscriber, Dict[str, List[Subscriber]]]) -> None:
    wi, s1, s2, subscriptions = workflow_instance
    assert wi.ack_for("start_collect", s1) is not None
    assert wi.ack_for("start_collect", s2) is not None
    assert wi.ack_for("collect_done", s1) is None
    assert wi.ack_for("collect_done", s2) is None


def test_pending_action_for(
    workflow_instance: Tuple[RunningTask, Subscriber, Subscriber, Dict[str, List[Subscriber]]],
) -> None:
    wi, s1, s2, subscriptions = workflow_instance
    # s1 already sent a done message for the current step
    assert wi.pending_action_for(s1) is None
    # s2 is still expected to provide a done message
    assert wi.pending_action_for(s2) == Action("collect", wi.id, "collect")


def test_handle_done(
    workflow_instance: Tuple[RunningTask, Subscriber, Subscriber, Dict[str, List[Subscriber]]]
) -> None:
    wi, s1, s2, subscriptions = workflow_instance
    # we are in state collect. Another ack of start is ignored.
    events = wi.handle_done(ActionDone("start", wi.id, "start", s1.id))  #
    assert wi.current_step.name == "collect"
    assert len(events) == 0
    # This step is unknown to the workflow and should not change its state
    events = wi.handle_done(ActionDone("boom", wi.id, "boom", s1.id))  #
    assert wi.current_step.name == "collect"
    assert len(events) == 0
    # this event has been received already and should not emit any new action
    events = wi.handle_done(ActionDone("collect", wi.id, "collect", s1.id))
    assert wi.current_step.name == "collect"
    assert len(events) == 0
    # this event is the last missing event in this step
    events = wi.handle_done(ActionDone("collect", wi.id, "collect", s2.id))
    assert wi.current_step.name == "done"
    assert len(events) == 1


def test_handle_error(
    workflow_instance: Tuple[RunningTask, Subscriber, Subscriber, Dict[str, List[Subscriber]]]
) -> None:
    wi, s1, s2, subscriptions = workflow_instance
    # this event is the last missing event in this step. It fails but the workflow should continue at that point
    events = wi.handle_error(ActionError("collect", wi.id, "collect", s2.id, "boom"))
    assert wi.current_step.name == "done"
    assert len(events) == 1
    # this step is configured to fail the whole workflow instance
    events = wi.handle_error(ActionError("collect_done", wi.id, "done", s2.id, "boom"))
    assert wi.is_active is False
    assert wi.is_error is True
    assert len(events) == 0


def test_complete_workflow(
    workflow_instance: Tuple[RunningTask, Subscriber, Subscriber, Dict[str, List[Subscriber]]],
    task_dependencies: TaskDependencies,
) -> None:
    init, s1, s2, subscriptions = workflow_instance
    # start new workflow instance
    wi, events = RunningTask.empty(
        init.descriptor, evolve(task_dependencies, subscribers_by_event=lambda: subscriptions)
    )
    assert wi.current_step.name == "start"
    assert len(events) == 2
    assert wi.progress.percentage == 0
    events = wi.handle_done(ActionDone("start", wi.id, "start", s1.id))
    assert wi.current_step.name == "wait"
    assert len(events) == 0
    assert wi.progress.percentage == 33
    handled, events = wi.handle_event(Event("godot", {"a": 2}))
    assert wi.current_step.name == "wait"
    assert handled is False
    assert len(events) == 0
    handled, events = wi.handle_event(Event("godot", {"a": 1, "d": "test"}))
    assert wi.current_step.name == "collect"
    assert handled is True
    assert len(events) == 2  # event from EmitEvent and action from PerformAction
    assert wi.progress.percentage == 33
    wi.handle_progress(ActionProgress("start", wi.id, "start", s1.id, ProgressDone("Test", 0, 100), utc()))
    wi.handle_progress(ActionProgress("start", wi.id, "start", s2.id, ProgressDone("Test", 50, 100), utc()))
    assert wi.progress.percentage == 50
    wi.handle_progress(ActionProgress("start", wi.id, "start", s1.id, ProgressDone("Test", 90, 100), utc()))
    wi.handle_progress(ActionProgress("start", wi.id, "start", s2.id, ProgressDone("Test", 80, 100), utc()))
    assert wi.progress.percentage == 60
    wi.handle_info(ActionInfo("start", wi.id, "start", s1.id, "error", "echt jetzt"))
    assert len(wi.info_messages) == 1
    events = wi.handle_done(ActionDone("start", wi.id, "start", s1.id))
    assert wi.current_step.name == "collect"
    assert len(events) == 0
    events = wi.handle_done(ActionDone("collect", wi.id, "collect", s1.id))
    assert wi.current_step.name == "collect"
    assert len(events) == 0
    events = wi.handle_done(ActionDone("collect", wi.id, "collect", s2.id))
    assert wi.current_step.name == "done"
    assert len(events) == 1
    assert wi.progress.percentage == 66
    events = wi.handle_done(ActionDone("done", wi.id, "done", s1.id))
    assert len(events) == 0
    assert wi.current_step.name == "done"
    events = wi.handle_done(ActionDone("done", wi.id, "done", s2.id))
    assert len(events) == 1
    assert wi.progress.percentage == 100
    assert wi.is_active is False


def test_marshalling_trigger() -> None:
    roundtrip(EventTrigger("test", {"foo": True}))
    roundtrip(TimeTrigger("* * * * *"))


def test_marshalling_step_action() -> None:
    roundtrip(PerformAction("test"))
    roundtrip(EmitEvent(Event("test", {"foo": "hello"})))
    roundtrip(WaitForEvent("test", {"foo": "hello"}))
    roundtrip(ExecuteCommand("help"))


def test_marshalling_task_command() -> None:
    roundtrip(SendMessage(Event("test", {"foo": "hello"})))
    roundtrip(ExecuteOnCLI("test", frozendict({"fii": "bla"})))


def test_marshalling_step() -> None:
    roundtrip(Step("test", PerformAction("test")))
    roundtrip(Step("test", WaitForEvent("test"), timedelta(seconds=19), StepErrorBehaviour.Stop))


def test_marshalling_job() -> None:
    j = Job(TaskDescriptorId("id"), ExecuteCommand("echo hello"), timedelta(seconds=10), EventTrigger("run_job"))
    roundtrip(j)
    roundtrip(Job(j.id, j.command, j.timeout, j.trigger, (EventTrigger("test"), timedelta(hours=2))))


def test_marshalling_workflow(test_workflow: Workflow) -> None:
    roundtrip(test_workflow)


def roundtrip(obj: Any) -> None:
    js = to_js(obj)
    again = from_js(js, type(obj))
    assert DeepDiff(obj, again) == {}, f"Json: {js} serialized as {again}"


def test_task_progress(task_dependencies: TaskDependencies) -> None:
    actions = ["collect", "encode"]
    wf = Workflow(TaskDescriptorId("test_workflow"), "name", [Step(a, PerformAction(a)) for a in actions], [])
    sb = Subscriber(SubscriberId("test"), {s: Subscription(s) for s in actions})
    rt, _ = RunningTask.empty(wf, evolve(task_dependencies, subscribers_by_event=lambda: {s: [sb] for s in actions}))

    def progress(step: str, fn: Callable[[int], ProgressDone]) -> None:
        # use revers order to test that the correct order is used
        for idx in range(3):
            rt.handle_progress(ActionProgress(step, rt.id, step, sb.id, fn(idx), utc()))

    # report progress start on the collect step
    progress("collect", lambda idx: ProgressDone(str(idx), 0, 100, path=["foo", "bla"]))
    assert [x["name"] for x in rt.progress_json()["parts"]] == ["0", "1", "2", "encode"]

    # report index as progress on the collect step
    progress("collect", lambda idx: ProgressDone(str(idx), idx, 100, path=["foo", "bla"]))
    assert [x["name"] for x in rt.progress_json()["parts"]] == ["2", "1", "0", "encode"]

    # report progress done on the collect step
    rt.handle_done(ActionDone("collect", rt.id, "collect", sb.id))
    assert [x["name"] for x in rt.progress_json()["parts"]] == ["collect", "encode"]
    assert rt.progress.overall_progress().percentage == 50

    # report progress done on the encode step
    progress("encode", lambda idx: ProgressDone(str(idx), 0, 100, path=["foo", "bla"]))
    assert [x["name"] for x in rt.progress_json()["parts"]] == ["collect", "0", "1", "2"]

    # report index as progress on the collect step
    progress("collect", lambda idx: ProgressDone(str(idx), idx, 100, path=["foo", "bla"]))
    assert [x["name"] for x in rt.progress_json()["parts"]] == ["collect", "2", "1", "0"]

    rt.handle_done(ActionDone("encode", rt.id, "encode", sb.id))
    assert [x["name"] for x in rt.progress_json()["parts"]] == ["collect", "encode"]
    assert rt.progress.overall_progress().percentage == 100


def test_emitting_task_progress(
    workflow_instance: Tuple[RunningTask, Subscriber, Subscriber, Dict[str, List[Subscriber]]]
) -> None:
    task, sub, _, _ = workflow_instance
    # after the task is created, there is not emitted progress
    assert task.not_emitted_progress() is not None
    # no progress update, nothing to emit
    assert task.not_emitted_progress() is None
    assert task.not_emitted_progress() is None
    # add progress updates
    task.handle_progress(ActionProgress("collect", task.id, "collect", sub.id, ProgressDone("collect", 0, 100), utc()))
    assert task.not_emitted_progress() is not None
    # no progress update, nothing to emit
    assert task.not_emitted_progress() is None
    assert task.not_emitted_progress() is None



import asyncio
from typing import List

import pytest
from arango.database import StandardDatabase

from resotocore.analytics import AnalyticsEventSender, InMemoryEventSender
from resotocore.db import modeldb
from resotocore.db.async_arangodb import AsyncArangoDB
from resotocore.db.entitydb import EventEntityDb
from resotocore.db.modeldb import ModelDb, EventModelDb
from resotocore.model.model import ComplexKind, Property, StringKind, NumberKind, BooleanKind, Kind


@pytest.fixture
def test_kinds() -> List[Kind]:
    string_kind = StringKind("some.string", 0, 3, "\\w+")
    int_kind = NumberKind("some.int", "int32", 0, 100)
    bool_kind = BooleanKind("some.bool")
    base = ComplexKind(
        "base",
        [],
        [
            Property("identifier", "string", required=True),
            Property("kind", "string", required=True),
        ],
    )
    foo = ComplexKind(
        "foo",
        ["base"],
        [
            Property("name", "string"),
            Property("some_int", "some.int"),
            Property("some_string", "some.string"),
            Property("now_is", "datetime"),
        ],
    )
    bla = ComplexKind(
        "bla",
        ["base"],
        [
            Property("name", "string"),
            Property("now", "date"),
            Property("f", "int32"),
            Property("g", "int32[]"),
        ],
    )
    return [string_kind, int_kind, bool_kind, base, foo, bla]


@pytest.fixture
async def model_db(test_db: StandardDatabase) -> ModelDb:
    async_db = AsyncArangoDB(test_db)
    model_db = modeldb.model_db(async_db, "model")
    await model_db.create_update_schema()
    await model_db.wipe()
    return model_db


@pytest.fixture
def event_db(model_db: ModelDb, event_sender: AnalyticsEventSender) -> EventModelDb:
    return EventEntityDb(model_db, event_sender, "model")


@pytest.mark.asyncio
async def test_load(model_db: ModelDb, test_kinds: List[Kind]) -> None:
    await model_db.update_many(test_kinds)
    loaded = [kind async for kind in model_db.all()]
    assert test_kinds.sort(key=fqn) == loaded.sort(key=fqn)


@pytest.mark.asyncio
async def test_update(model_db: ModelDb, test_kinds: List[Kind]) -> None:
    # multiple updates should work as expected
    await model_db.update_many(test_kinds)
    await model_db.update_many(test_kinds)
    await model_db.update_many(test_kinds)
    loaded = [kind async for kind in model_db.all()]
    assert test_kinds.sort(key=fqn) == loaded.sort(key=fqn)


@pytest.mark.asyncio
async def test_delete(model_db: ModelDb, test_kinds: List[Kind]) -> None:
    await model_db.update_many(test_kinds)
    remaining = list(test_kinds)
    for _ in test_kinds:
        kind = remaining.pop()
        await model_db.delete_value(kind)
        loaded = [kind async for kind in model_db.all()]
        assert remaining.sort(key=fqn) == loaded.sort(key=fqn)
    assert len([kind async for kind in model_db.all()]) == 0


@pytest.mark.asyncio
async def test_events(event_db: EventModelDb, test_kinds: List[Kind], event_sender: InMemoryEventSender) -> None:
    # 2 times update
    await event_db.update_many(test_kinds)
    await event_db.update_many(test_kinds)
    # 6 times delete
    for kind in test_kinds:
        await event_db.delete_value(kind)
    # make sure all events will arrive
    await asyncio.sleep(0.1)
    # ensure the correct count and order of events
    assert [a.kind for a in event_sender.events] == ["model-updated-many"] * 2 + ["model-deleted"] * 6


def fqn(kind: Kind) -> str:
    return kind.fqn



import re
from datetime import datetime

from typing import Optional, Union, Pattern, Any

from resotocore.model.adjust_node import DirectAdjuster, NoAdjust
from resotocore.types import Json
from resotocore.util import utc_str, value_in_path


def test_adjust_expired() -> None:
    adjuster = DirectAdjuster()
    created_at = datetime(2021, 1, 1)
    expires_at = datetime(2022, 2, 15, 13)

    def expect_expires(reported: Json, expires: Optional[Union[str, Pattern[Any]]]) -> None:
        result = adjuster.adjust({"reported": reported})
        if expires is None:
            assert value_in_path(result, ["metadata", "expires"]) is None
        elif isinstance(expires, str):
            assert value_in_path(result, ["metadata", "expires"]) == expires
        elif isinstance(expires, Pattern):
            value = value_in_path(result, ["metadata", "expires"])
            assert value is not None
            assert expires.fullmatch(value)

    # test iso datetime
    expect_expires({"tags": {"resoto:expires": utc_str(expires_at)}}, "2022-02-15T13:00:00Z")
    expect_expires({"tags": {"resoto:expires": expires_at.isoformat()}}, re.compile("2022-02-1.*"))
    expect_expires({"tags": {"resoto:expires": "15.02.2022"}}, re.compile("2022-02-1.*"))
    expect_expires({"tags": {"resoto:expires": "02/15/2022"}}, re.compile("2022-02-1.*"))
    expect_expires({"tags": {"resoto:expires": "+2d"}}, None)  # ony accept absolute time
    expect_expires({"tags": {"resoto:expires": "never"}}, None)  # never can not be parsed

    # test duration
    reported: Json = {"ctime": utc_str(created_at)}
    expect_expires({**reported, "tags": {"expiration": "never"}}, None)  # never can not be parsed
    expect_expires({**reported, "tags": {"resoto:expiration": "never"}}, None)  # never can not be parsed
    expect_expires({"tags": {"resoto:expiration": "12h"}}, None)  # no ctime given
    expect_expires({"tags": {"expiration": "2mo3d4h5min"}}, None)  # no ctime given
    expect_expires({**reported, "tags": {"resoto:expiration": "2mo3d4h5min"}}, "2021-03-07T04:05:00Z")
    expect_expires({**reported, "tags": {"expiration": "2mo3d4h5min"}}, "2021-03-07T04:05:00Z")

    # multiple values given: use order: ck:expiration -> ck:expires -> expiration -> expires
    expect_expires(
        {
            **reported,
            "tags": {
                "expiration": "2d",
                "resoto:expires": "2021-01-01T11:20:00Z",
                "resoto:expiration": "2w3d4h5m",
            },
        },
        "2021-01-01T11:20:00Z",
    )

    # no tags given
    expect_expires({}, None)


def test_no_adjust() -> None:
    adjuster = NoAdjust()
    created_at = datetime(2021, 1, 1)
    expires_at = datetime(2022, 2, 1)
    reported: Json = {"ctime": utc_str(created_at)}

    def expect_expires(reported: Json, expires: Optional[str]) -> None:
        result = adjuster.adjust({"reported": reported})
        assert value_in_path(result, ["metadata", "expires"]) == expires

    expect_expires({"tags": {"expires": utc_str(expires_at)}}, None)
    expect_expires({"tags": {"resoto:expires": "2w3d4h5m"}, **reported}, None)



from __future__ import annotations

import asyncio
import logging
import re
from asyncio import Task, CancelledError
from contextlib import suppress
from copy import copy
from datetime import timedelta
from typing import Optional, Any, Callable, Union, Sequence, Dict, List, Tuple
from aiostream import stream
from attrs import evolve

from resotocore.analytics import AnalyticsEventSender, CoreEvent
from resotocore.cli.model import CLIContext, CLI
from resotocore.core_config import CoreConfig
from resotocore.db.jobdb import JobDb
from resotocore.db.runningtaskdb import RunningTaskData, RunningTaskDb
from resotocore.ids import SubscriberId, TaskDescriptorId
from resotocore.message_bus import (
    MessageBus,
    Event,
    Action,
    ActionDone,
    Message,
    ActionError,
    ActionInfo,
    ActionProgress,
    CoreMessage,
)
from resotocore.model.db_updater import GraphMerger
from resotocore.service import Service
from resotocore.task import TaskHandler, RunningTaskInfo
from resotocore.task.model import Subscriber
from resotocore.task.scheduler import Scheduler
from resotocore.task.start_workflow_on_first_subscriber import wait_and_start
from resotocore.task.subscribers import SubscriptionHandler
from resotocore.task.task_dependencies import TaskDependencies
from resotocore.task.task_description import (
    Workflow,
    RunningTask,
    EventTrigger,
    TimeTrigger,
    TaskSurpassBehaviour,
    PerformAction,
    Step,
    TaskDescription,
    Job,
    ExecuteCommand,
    TaskCommand,
    SendMessage,
    ExecuteOnCLI,
    StepErrorBehaviour,
    RestartAgainStepAction,
    Trigger,
    WaitForCollectDone,
)
from resotocore.util import first, Periodic, group_by, utc_str, utc, partition_by
from resotocore.types import Json

log = logging.getLogger(__name__)


class TaskHandlerService(TaskHandler, Service):
    # region init

    def __init__(
        self,
        running_task_db: RunningTaskDb,
        job_db: JobDb,
        message_bus: MessageBus,
        event_sender: AnalyticsEventSender,
        subscription_handler: SubscriptionHandler,
        graph_merger: GraphMerger,
        scheduler: Scheduler,
        cli: CLI,
        config: CoreConfig,
    ):
        self.running_task_db = running_task_db
        self.job_db = job_db
        self.message_bus = message_bus
        self.event_sender = event_sender
        self.subscription_handler = subscription_handler
        self.scheduler = scheduler
        self.cli = cli
        self.cli_context = CLIContext(source="task_handler")
        self.config = config
        # note: the waiting queue is kept in memory and lost when the service is restarted.
        self.start_when_done: Dict[str, TaskDescription] = {}

        # Step1: define all workflows and jobs in code: later it will be persisted and read from database
        self.task_descriptions: Sequence[TaskDescription] = [*self.known_workflows(config), *self.known_jobs()]
        self.tasks: Dict[str, RunningTask] = {}
        self.message_bus_watcher: Optional[Task[None]] = None
        self.initial_start_workflow_task: Optional[Task[None]] = None
        self.timeout_watcher = Periodic("task_watcher", self.check_running_tasks, timedelta(seconds=10))
        self.registered_event_trigger: List[Tuple[EventTrigger, TaskDescription]] = []
        self.registered_event_trigger_by_message_type: Dict[str, List[Tuple[EventTrigger, TaskDescription]]] = {}
        self.task_dependencies = TaskDependencies(graph_merger, subscription_handler.subscribers_by_event)

    # endregion

    # region startup and teardown

    async def update_trigger(self, desc: TaskDescription, register: bool = True) -> None:
        # safeguard: unregister all event trigger of this task
        for existing in (tup for tup in self.registered_event_trigger if desc.id == tup[1].id):
            self.registered_event_trigger.remove(existing)
        # safeguard: unregister all schedule trigger of this task
        for job in self.scheduler.list_jobs():
            if str(job.id).startswith(desc.id):
                job.remove()
        # add all triggers
        if register:
            for trigger in desc.triggers:
                if isinstance(trigger, EventTrigger):
                    self.registered_event_trigger.append((trigger, desc))
                if isinstance(trigger, TimeTrigger):
                    uid = f"{desc.id}_{trigger.cron_expression}"
                    name = f"Trigger for task {desc.id} on cron expression {trigger.cron_expression}"
                    self.scheduler.cron(uid, name, trigger.cron_expression, self.time_triggered, desc, trigger)
        # recompute the lookup table
        self.registered_event_trigger_by_message_type = group_by(
            lambda t: t[0].message_type, self.registered_event_trigger
        )

    # task descriptors can hold placeholders (e.g. @NOW@)
    # which should be replaced, when the task is started (or restarted).
    def evaluate_task_definition(self, descriptor: TaskDescription, **env: str) -> TaskDescription:
        def evaluate(step: Step) -> Step:
            if isinstance(step.action, ExecuteCommand):
                update = copy(step)
                update.action = ExecuteCommand(self.cli.replace_placeholder(step.action.command, **env))
                return update
            else:
                return step

        updated = copy(descriptor)
        updated.steps = [evaluate(step) for step in descriptor.steps]
        return updated

    async def fetch_task_metadata(self, descriptor_id: TaskDescriptorId) -> Json:
        timing: Dict[str, Dict[str, int]] = {}
        if last_workflow_run := await self.running_task_db.last(descriptor_id=descriptor_id):
            timing = {
                s.step_name: {
                    "started_at": int(s.started_at.timestamp()),
                    "finished_at": int(s.finished_at.timestamp()),
                }
                for s in last_workflow_run.step_states
                if s.finished_at and s.started_at
            }
        return {
            "timing": timing,
        }

    async def start_task_directly(self, desc: TaskDescription, reason: str) -> RunningTask:
        updated = self.evaluate_task_definition(desc)

        metadata = await self.fetch_task_metadata(desc.id)
        task, commands = RunningTask.empty(updated, self.task_dependencies, metadata=metadata)

        log.info(f"Start new task: {updated.name} with id {task.id}")
        # store initial state in database
        await self.running_task_db.insert(task)
        self.tasks[task.id] = task
        await self.execute_task_commands(task, commands)
        await self.event_sender.core_event(
            CoreEvent.TaskStarted,
            {
                "reason": reason,
                "task_descriptor_id": updated.id,
                "task_descriptor_name": updated.name,
                "kind": type(updated).__name__,
            },
        )
        return task

    async def start_task(self, desc: TaskDescription, reason: str) -> Optional[RunningTaskInfo]:
        existing = first(lambda x: x.descriptor.id == desc.id and x.is_active, self.tasks.values())
        if existing:
            if desc.on_surpass == TaskSurpassBehaviour.Skip:
                log.info(
                    f"Task {desc.name} has been triggered. Since the last job is not finished, "
                    f"the execution will be skipped, as defined by the task"
                )
                return None
            elif desc.on_surpass == TaskSurpassBehaviour.Replace:
                log.info(f"New task {desc.name} should replace existing run: {existing.id}.")
                existing.end()
                await self.store_running_task_state(existing)
                return RunningTaskInfo(await self.start_task_directly(desc, reason))
            elif desc.on_surpass == TaskSurpassBehaviour.Parallel:
                log.info(f"New task {desc.name} will race with existing run {existing.id}.")
                return RunningTaskInfo(await self.start_task_directly(desc, reason))
            elif desc.on_surpass == TaskSurpassBehaviour.Wait:
                log.info(f"New task {desc.name} with reason {reason} will run when existing run {existing.id} is done.")
                self.start_when_done[desc.id] = desc
                return RunningTaskInfo(existing, True)
            else:
                raise AttributeError(f"Surpass behaviour not handled: {desc.on_surpass}")
        else:
            return RunningTaskInfo(await self.start_task_directly(desc, reason))

    async def start_interrupted_tasks(self) -> List[RunningTask]:
        descriptions = {w.id: w for w in self.task_descriptions}

        def reset_state(wi: RunningTask, task_data: RunningTaskData) -> RunningTask:
            infos, messages = partition_by(lambda x: isinstance(x, ActionInfo), task_data.received_messages)
            # reset the received messages
            wi.received_messages = messages
            wi.info_messages = infos  # type: ignore
            # move the fsm into the last known state
            wi.machine.set_state(task_data.current_state_name)
            # import state of the current step
            wi.current_state.import_state(task_data.current_state_snapshot)
            for si in task_data.step_states:
                if step_state := wi.states.get(si.step_name):
                    step_state.started_at = si.started_at
                    step_state.finished_at = si.finished_at
                    step_state.timed_out = si.timed_out
            # reset times
            wi.task_started_at = task_data.task_started_at
            # ignore all messages that would be emitted
            wi.move_to_next_state()
            return wi

        instances: List[RunningTask] = []
        async for data in self.running_task_db.all_running():
            descriptor = descriptions.get(data.task_descriptor_id)
            if descriptor:
                # we have captured the timestamp when the task has been started
                updated = self.evaluate_task_definition(descriptor, now=utc_str(data.task_started_at))
                rt = RunningTask(data.id, updated, self.task_dependencies)
                instance = reset_state(rt, data)
                if isinstance(instance.current_step.action, RestartAgainStepAction):
                    log.info(f"Restart interrupted action: {instance.current_step.action}")
                    await self.execute_task_commands(instance, instance.current_state.commands_to_execute())
                instances.append(instance)

            else:
                log.warning(f"No task description with this id found: {data.task_descriptor_id}. Remove instance data.")
                await self.running_task_db.delete(data.id)
        return instances

    async def __aenter__(self) -> TaskHandlerService:
        return await self.start()

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        return await self.stop()

    async def start(self) -> TaskHandlerService:
        log.debug("TaskHandlerService is starting up!")

        # load job descriptions from database
        db_jobs = [job async for job in self.job_db.all()]
        self.task_descriptions = [*self.task_descriptions, *db_jobs]

        # load and restore all tasks
        self.tasks = {wi.id: wi for wi in await self.start_interrupted_tasks()}

        await self.timeout_watcher.start()

        for descriptor in self.task_descriptions:
            await self.update_trigger(descriptor)

        if self.config.runtime.start_collect_on_subscriber_connect:
            filtered = [wf for wf in self.known_workflows(self.config) if wf.id == "collect_and_cleanup"]
            self.initial_start_workflow_task = wait_and_start(filtered, self, self.message_bus)

        async def listen_to_message_bus() -> None:
            async with self.message_bus.subscribe(SubscriberId("resotocore.task_handler")) as messages:
                while True:
                    message = None
                    try:
                        message = await messages.get()
                        if isinstance(message, Event):
                            await self.handle_event(message)
                        elif isinstance(message, Action):
                            await self.handle_action(message)
                        elif isinstance(message, (ActionDone, ActionError)):
                            log.info(f"Ignore message via event bus: {message}")
                    except asyncio.CancelledError as ex:
                        # if we outer task is cancelled, give up
                        raise ex
                    except Exception as ex:
                        log.error(f"Could not handle event {message} - give up.", exc_info=ex)

        self.message_bus_watcher = asyncio.create_task(listen_to_message_bus())
        return self

    async def stop(self) -> None:
        log.info("Tear down task handler")
        # deregister from all triggers
        for descriptor in self.task_descriptions:
            await self.update_trigger(descriptor, register=False)

        # stop timeout watcher
        await self.timeout_watcher.stop()

        # stop event listener
        if self.message_bus_watcher:
            self.message_bus_watcher.cancel()
            try:
                await self.message_bus_watcher
            except CancelledError:
                log.info("task has been cancelled")

        # wait for all running commands to complete
        for task in list(self.tasks.values()):
            if task.update_task:
                with suppress(Exception):
                    await task.update_task
                del self.tasks[task.id]

        # in case the task is not done
        if self.initial_start_workflow_task and not self.initial_start_workflow_task.done():
            self.initial_start_workflow_task.cancel()

    # endregion

    # region job handler

    async def running_tasks(self) -> List[RunningTask]:
        return list(self.tasks.values())

    async def start_task_by_descriptor_id(self, uid: TaskDescriptorId) -> Optional[RunningTaskInfo]:
        td = first(lambda t: t.id == uid, self.task_descriptions)
        if td:
            return await self.start_task(td, "direct")
        else:
            raise NameError(f"No task with such id: {uid}")

    async def list_jobs(self) -> List[Job]:
        return [td for td in self.task_descriptions if isinstance(td, Job)]

    async def list_workflows(self) -> List[Workflow]:
        return [td for td in self.task_descriptions if isinstance(td, Workflow)]

    def __check_system_job(self, name: str) -> None:
        # : is not allowed in job names
        if re.match(r"^[^:]+$", name) is None:
            raise AttributeError(f"System jobs can't be modified: {name}")

    async def add_job(self, job: Job, force: bool = False) -> None:
        if not force:
            self.__check_system_job(job.name)
        descriptions = list(self.task_descriptions)
        existing = first(lambda td: td.id == job.id, descriptions)
        if existing:
            if not existing.mutable:
                raise AttributeError(f"There is an existing job with this {job.id} which can not be deleted!")
            log.info(f"Job with id {job.id} already exists. Update this job.")
            descriptions.remove(existing)
        # store in database
        await self.job_db.update(job)
        descriptions.append(job)
        self.task_descriptions = descriptions
        await self.update_trigger(job)

    async def delete_running_task(self, task: RunningTask) -> None:
        # send analytics event
        await self.event_sender.core_event(
            CoreEvent.TaskCompleted,
            {
                "task_descriptor_id": task.descriptor.id,
                "task_descriptor_name": task.descriptor.name,
                "kind": type(task.descriptor).__name__,
                "success": not task.is_error,
            },
            duration=(utc() - task.task_started_at).total_seconds(),
            step_count=len(task.descriptor.steps),
        )
        task.descriptor_alive = False
        # remove tasks from list of running tasks
        self.tasks.pop(task.id, None)
        if task.update_task and not task.update_task.done():
            task.update_task.cancel()

        # mark step as error
        task.end()
        await self.mark_done_in_database(task)

    async def delete_job(self, job_id: str, force: bool = False) -> Optional[Job]:
        if not force:
            self.__check_system_job(job_id)
        job: Job = first(lambda td: td.id == job_id and isinstance(td, Job), self.task_descriptions)  # type: ignore
        if job:
            if not job.mutable:
                raise AttributeError(f"Can not delete job: {job.id} - it is defined in a system file!")
            # delete all running tasks of this job
            for task in list(filter(lambda x: x.descriptor.id == job.id, self.tasks.values())):
                log.info(f"Job: {job_id}: delete running task: {task.id}")
                await self.delete_running_task(task)
            await self.job_db.delete(job_id)
            descriptions = list(self.task_descriptions)
            descriptions.remove(job)
            self.task_descriptions = descriptions
            await self.update_trigger(job, register=False)
        return job

    # endregion

    # region maintain running tasks

    async def time_triggered(self, descriptor: TaskDescription, trigger: TimeTrigger) -> None:
        log.info(f"Task {descriptor.name} triggered by time: {trigger.cron_expression}")
        await self.start_task(descriptor, "time")

    async def check_for_task_to_start_on_message(self, msg: Message) -> None:
        # check if this event triggers any new task
        for trigger, descriptor in self.registered_event_trigger_by_message_type.get(msg.message_type, []):
            if msg.message_type == trigger.message_type:
                comp = trigger.filter_data
                if {key: msg.data.get(key) for key in comp} == comp if comp else True:
                    log.info(f"Event {msg.message_type} triggers task: {descriptor.name}")
                    await self.start_task(descriptor, "event")

    async def handle_event(self, event: Event) -> None:
        # check if any running task want's to handle this event
        for wi in list(self.tasks.values()):
            handled, commands = wi.handle_event(event)
            if handled:
                await self.execute_task_commands(wi, commands, event)

        # check if this event triggers any new task
        await self.check_for_task_to_start_on_message(event)

    # noinspection PyMethodMayBeStatic
    async def handle_action(self, action: Action) -> None:
        await self.check_for_task_to_start_on_message(action)

    async def handle_action_result(
        self, done: Union[ActionDone, ActionError], fn: Callable[[RunningTask], Sequence[TaskCommand]]
    ) -> None:
        wi = self.tasks.get(done.task_id)
        if wi:
            commands = fn(wi)
            await self.execute_task_commands(wi, commands, done)
        else:
            log.warning(
                f"Received an ack for an unknown task={done.task_id} "
                f"event={done.message_type} from={done.subscriber_id}. Ignore."
            )

    async def handle_action_done(self, done: ActionDone) -> None:
        return await self.handle_action_result(done, lambda wi: wi.handle_done(done))

    async def handle_action_error(self, err: ActionError) -> None:
        log.info(f"Received error: {err.error} {err.step_name}:{err.message_type} of {err.task_id}")
        if rt := self.tasks.get(err.task_id):
            await self.handle_action_result(err, lambda wi: wi.handle_error(err))
            rt.info_messages.append(err)
            name = rt.descriptor.name
            await self.event_sender.core_event(CoreEvent.ActionError, {"workflow": name, "message": err.error})
            await self.message_bus.emit_event(
                CoreMessage.ErrorMessage, {"workflow": name, "task": rt.id, "message": err.error}
            )

    async def handle_action_info(self, info: ActionInfo) -> None:
        if rt := self.tasks.get(info.task_id):
            rt.handle_info(info)
            name = rt.descriptor.name
            await self.running_task_db.update_state(rt, info)
            await self.event_sender.core_event(CoreEvent.ActionError, {"workflow": name, "message": info.message})
            if info.level == "error":
                await self.message_bus.emit_event(
                    CoreMessage.ErrorMessage, {"workflow": name, "task": rt.id, "message": info.message}
                )

    async def handle_action_progress(self, info: ActionProgress) -> None:
        if rt := self.tasks.get(info.task_id):  # get the related running task
            if info.step_name == rt.current_step.name:  # make sure this progress belongs to the current step
                log.debug("Received progress: %s", info)
                rt.handle_progress(info)

    async def execute_task_commands(
        self, wi: RunningTask, commands: Sequence[TaskCommand], origin_message: Optional[Message] = None
    ) -> None:
        async def execute_commands() -> None:
            # execute and collect all task commands
            results: Dict[TaskCommand, Any] = {}
            for command in commands:
                try:
                    if isinstance(command, SendMessage):
                        await self.message_bus.emit(command.message)
                        results[command] = None
                    elif isinstance(command, ExecuteOnCLI):
                        ctx = evolve(self.cli_context, env={**command.env, **wi.descriptor.environment})
                        result = await self.cli.execute_cli_command(command.command, stream.list, ctx)
                        results[command] = result
                    else:
                        raise AttributeError(f"Does not understand this command: {wi.descriptor.name}:  {command}")
                except Exception as ex:
                    log.info(f"Error executing command: {command} {ex}")
                    results[command] = ex

            # The descriptor might be removed in the meantime. If this is the case stop execution.
            if wi.descriptor_alive:
                active_before_result = wi.is_active
                # before we move on, we need to store the current state of the task (or delete if it is done)
                await self.store_running_task_state(wi, origin_message)
                # inform the task about the result, which might trigger new tasks to execute
                new_commands = wi.handle_command_results(results)
                if new_commands:
                    # note: recursion depth is defined by the number of steps in a job description and should be safe.
                    await self.execute_task_commands(wi, new_commands)
                elif active_before_result and not wi.is_active:
                    # if this was the last result the task was waiting for, delete the task
                    await self.store_running_task_state(wi, origin_message)

        async def execute_in_order(task: Task[Any]) -> None:
            # make sure the last execution is finished, before the new execution starts
            await task
            await execute_commands()

        # start execution of commands in own task to not block the task handler
        # note: the task is awaited finally in the timeout handler or context handler shutdown
        wi.update_task = asyncio.create_task(execute_in_order(wi.update_task) if wi.update_task else execute_commands())

    async def store_running_task_state(self, wi: RunningTask, origin_message: Optional[Message] = None) -> None:
        if wi.is_active:
            await self.running_task_db.update_state(wi, origin_message)
        elif wi.id in self.tasks:
            # only here as safeguard: in case of a core restart, the task would be restarted again
            # cleanup happens in the overdue handler
            await self.mark_done_in_database(wi)

    async def list_all_pending_actions_for(self, subscriber: Subscriber) -> List[Action]:
        pending = map(lambda x: x.pending_action_for(subscriber), self.tasks.values())
        return [x for x in pending if x]

    async def mark_done_in_database(self, task: RunningTask) -> None:
        # workflows are kept in database, jobs are deleted
        with suppress(Exception):
            if isinstance(task.descriptor, Job):
                await self.running_task_db.delete(task.id)
            else:
                await self.running_task_db.update_state(task)

    # endregion

    # region periodic task checker

    async def check_running_tasks(self) -> None:
        """
        Called periodically by the system.
        In case there is an overdue task, an action error is injected into the task.
        In case there is progress not emitted, it is send to the message bus.
        """
        for task in list(self.tasks.values()):
            # check if there is any pending progress update to emit
            if task.not_emitted_progress() is not None:
                await self.message_bus.emit_event(
                    CoreMessage.ProgressMessage,
                    {"workflow": task.descriptor.name, "task": task.id, "message": task.progress_json()},
                )

            # check if active tasks are overdue
            if task.is_active:
                if task.current_state.check_timeout():
                    if task.current_step.on_error == StepErrorBehaviour.Continue:
                        current_step = task.current_step.name
                        commands = task.move_to_next_state()
                        log.warning(
                            f"Task {task.id}: {task.descriptor.name} timed out in step "
                            f"{current_step}. Moving on to step: {task.current_step.name}."
                        )
                        await self.execute_task_commands(task, commands)
                    else:
                        log.warning(
                            f"Task {task.id}: {task.descriptor.name} timed out "
                            f"in step {task.current_step.name}. Stop the task."
                        )
                        task.end()
                        await self.store_running_task_state(task)
            # check again for active (might have changed for overdue tasks)
            if not task.is_active:
                if task.update_task and not task.update_task.cancelled():
                    with suppress(Exception, CancelledError):
                        await task.update_task
                await self.delete_running_task(task)
            # if the task is finished, check if there is already the next run to start
            if task.id not in self.tasks and task.descriptor.id in self.start_when_done:
                self.start_when_done.pop(task.descriptor.id, None)
                await self.start_task_directly(task.descriptor, "previous_task_finished")

    # endregion

    # region known task descriptors

    @staticmethod
    def known_jobs() -> List[Job]:
        return []

    @staticmethod
    def known_workflows(config: CoreConfig) -> List[Workflow]:
        def workflow(name: TaskDescriptorId, steps: List[Step]) -> Workflow:
            trigger: List[Trigger] = [EventTrigger(f"start_{name}_workflow")]
            wf_config = config.workflows.get(name)
            if wf_config:
                trigger.append(TimeTrigger(wf_config.schedule))
            return Workflow(uid=name, name=name, steps=steps, triggers=trigger, on_surpass=TaskSurpassBehaviour.Wait)

        collect_steps = [
            Step("pre_collect", PerformAction("pre_collect"), timedelta(seconds=10)),
            Step("collect", PerformAction("collect"), timedelta(seconds=10)),
            Step("wait_for_graph_merged", WaitForCollectDone(), timedelta(minutes=10)),
            Step("merge_outer_edges", PerformAction("merge_outer_edges"), timedelta(seconds=10)),
            Step("post_collect", PerformAction("post_collect"), timedelta(seconds=10)),
        ]
        cleanup_steps = [
            Step("pre_cleanup_plan", PerformAction("pre_cleanup_plan"), timedelta(seconds=10)),
            Step("cleanup_plan", PerformAction("cleanup_plan"), timedelta(seconds=10)),
            Step("post_cleanup_plan", PerformAction("post_cleanup_plan"), timedelta(seconds=10)),
            Step("pre_cleanup", PerformAction("pre_cleanup"), timedelta(seconds=10)),
            Step("cleanup", PerformAction("cleanup"), timedelta(seconds=10)),
            Step("post_cleanup", PerformAction("post_cleanup"), timedelta(seconds=10)),
        ]
        metrics_steps = [
            Step("pre_generate_metrics", PerformAction("pre_generate_metrics"), timedelta(seconds=10)),
            Step("generate_metrics", PerformAction("generate_metrics"), timedelta(seconds=10)),
            Step("post_generate_metrics", PerformAction("post_generate_metrics"), timedelta(seconds=10)),
        ]
        return [
            workflow(TaskDescriptorId("collect"), collect_steps + metrics_steps),
            workflow(TaskDescriptorId("cleanup"), cleanup_steps + metrics_steps),
            workflow(TaskDescriptorId("metrics"), metrics_steps),
            workflow(TaskDescriptorId("collect_and_cleanup"), collect_steps + cleanup_steps + metrics_steps),
        ]

    # endregion



from datetime import datetime, timedelta
from functools import partial
from typing import Callable, Optional, Any, List

import pytest
from attrs import evolve
from deepdiff import DeepDiff
from hypothesis import given, settings, HealthCheck
from parsy import Parser, ParseError
from pytest import approx

from resotocore import error
from resotocore.model.graph_access import EdgeTypes, Direction
from resotocore.query.model import (
    Navigation,
    Part,
    Query,
    P,
    AggregateVariable,
    Aggregate,
    AggregateFunction,
    IsTerm,
    IdTerm,
    Sort,
    WithClause,
    WithClauseFilter,
    AggregateVariableName,
    AggregateVariableCombined,
    AllTerm,
    MergeTerm,
    MergeQuery,
    FulltextTerm,
    CombinedTerm,
    Predicate,
    Limit,
    NotTerm,
    WithUsage,
)
from resotocore.query.query_parser import (
    predicate_term,
    is_term,
    function_term,
    combined_term,
    navigation_parser,
    part_parser,
    query_parser,
    preamble_parser,
    preamble_tags_parser,
    aggregate_group_variable_parser,
    aggregate_group_function_parser,
    aggregate_parser,
    id_term,
    sort_parser,
    limit_parser,
    with_clause_parser,
    not_term,
    term_parser,
    parse_query,
    context_term,
    with_usage_parser,
)
from resotocore.util import utc, parse_utc
from tests.resotocore.query import query


def test_parse_is_term() -> None:
    assert is_term.parse("is(test)") == IsTerm(["test"])
    assert is_term.parse("is(a, b, c)") == IsTerm(["a", "b", "c"])
    assert is_term.parse("is([a,b,c])") == IsTerm(["a", "b", "c"])
    assert is_term.parse("is(volume, instance)") == IsTerm(["volume", "instance"])
    assert_round_trip(is_term, IsTerm(["volume", "instance"]))
    assert_round_trip(is_term, IsTerm(["test"]))


def test_parse_id_term() -> None:
    assert id_term.parse('id("test")') == IdTerm("test")
    assert id_term.parse("id(test)") == IdTerm("test")


def test_parse_predicate() -> None:
    assert_round_trip(predicate_term, P("mem") < 23)
    assert_round_trip(predicate_term, P("simple") == "hallo")
    assert_round_trip(predicate_term, P("simple") != "hallo")
    assert_round_trip(predicate_term, P("simple").matches("^some.regex[a-d]+$"))
    assert_round_trip(predicate_term, P("simple").not_matches("^some.regex[a-d]+$"))
    assert_round_trip(predicate_term, P("num") > 23)
    assert_round_trip(predicate_term, P("num") >= 23)
    assert_round_trip(predicate_term, P("num") == 23)
    assert_round_trip(predicate_term, P("num") <= 23)
    assert_round_trip(predicate_term, P("num") < 23)
    assert_round_trip(predicate_term, P("num").is_in([1, 2, 5]))
    assert_round_trip(predicate_term, P("num").is_not_in([1, 2, 5]))


def test_parse_predicate_array() -> None:
    assert_round_trip(predicate_term, P.array("mem").for_any() < 23)
    assert_round_trip(predicate_term, P.array("mem").for_all() >= 23)
    assert_round_trip(predicate_term, P.array("mem").for_none().matches("foo.*"))
    assert_round_trip(predicate_term, P.array("num").for_any().is_in([1, 2, 5]))
    assert_round_trip(predicate_term, P.array("num").for_all().is_in([1, 2, 5]))
    assert_round_trip(predicate_term, P.array("num").for_none().is_in([1, 2, 5]))


def test_parse_context() -> None:
    assert context_term.parse("a.b[*].{ c<1 and d>2 }") == P.context("a.b[*]", P.single("c") < 1, P.single("d") > 2)
    assert_element = partial(assert_round_trip, context_term)
    assert_element(P.context("foo", P.single("mem") < 23))
    assert_element(P.context("foo", P.array("mem").for_all() < 23, P.single("cpu") > 2))
    assert_element(P.context("foo", P.single("core") > 2, P.context("inner", P.array("c").for_any() < 23)))


def test_kind() -> None:
    assert_round_trip(is_term, P.of_kind("foo"))


def test_function() -> None:
    assert_round_trip(function_term, P.function("in_subnet").on("foo.bla.bar", 1, "2", True))
    assert_round_trip(function_term, P.function("in_subnet").on("foo.bla.bar", "in_subnet"))
    assert_round_trip(function_term, P.function("in_subnet").on("foo.bla.bar", "in_subnet", "1000"))
    assert_round_trip(function_term, P.function("has_key").on("foo.bla.bar", "a", "b", "c"))
    assert_round_trip(function_term, P.function("has_key").on("foo.bla.bar", "a"))


def test_combined() -> None:
    assert_round_trip(combined_term, P.of_kind("foo") | P.of_kind("bla"))


def test_not() -> None:
    assert_round_trip(not_term, (P.with_id("foo") | P.of_kind("bla")).not_term())
    assert_round_trip(not_term, P.of_kind("bla").not_term())
    assert_round_trip(not_term, term_parser.parse("not(is(a) or not is(b) and not a>1 or not b<2 or not(a>1))"))
    # make sure not only negates the simple term, not the combined term
    assert term_parser.parse("not a==b and b==c") == CombinedTerm(NotTerm(P("a") == "b"), "and", P("b") == "c")


def test_filter_term() -> None:
    assert_round_trip(term_parser, P("mem") < 23)
    assert_round_trip(term_parser, P.with_id("foo"))
    assert_round_trip(term_parser, P.of_kind("foo"))
    assert_round_trip(term_parser, P.of_kind("foo") | P.of_kind("bla"))
    assert_round_trip(
        term_parser,
        ((P.of_kind("foo") | P.of_kind("bla")) & (P("a") > 23)) & (P("b").is_in([1, 2, 3])) & (P("c") == {"a": 123}),
    )


def test_fulltext_term() -> None:
    assert_round_trip(term_parser, FulltextTerm("test"))
    assert term_parser.parse('"foo"') == FulltextTerm("foo")
    # multiple strings are not allowed
    with pytest.raises(ParseError):
        term_parser.parse("foo bla bar")
    # multiple strings in quotes are allowed
    assert term_parser.parse('"foo bla bar"') == FulltextTerm("foo bla bar")
    # combined term can be parsed
    assert term_parser.parse('"foo" and test>3') == CombinedTerm(
        FulltextTerm("foo"), "and", Predicate("test", ">", 3, {})
    )
    assert term_parser.parse('a>1 and ("b" and (c<1 or "d") and "e") or "f" and g==2')


def test_merge_term() -> None:
    next_foo = Query.by(AllTerm()).traverse_in(until=Navigation.Max).filter("foo")
    query = Query.by(MergeTerm(Query.mk_term("bla"), [MergeQuery("foo123", next_foo)], Query.mk_term("bla")))
    assert_round_trip(term_parser, query)


def test_navigation_default() -> None:
    assert str(Navigation(1, 1)) == "-default->"
    assert str(Navigation(1, Navigation.Max)) == "-default[1:]->"


def test_navigation() -> None:
    for edge_type in EdgeTypes.all:
        # the default edge type is not rendered, so we set it explicitly to make the mapping homogeneous
        for start, until in [(0, 0), (1, 1), (5, 5), (1, 10), (1, Navigation.Max), (10, Navigation.Max)]:
            assert_round_trip(navigation_parser, Navigation(start, until, [edge_type], Direction.any, [edge_type]))
            for direction in Direction.all:
                assert_round_trip(navigation_parser, Navigation(start, until, [edge_type], direction))


def test_part() -> None:
    assert_round_trip(part_parser, Part(P.of_kind("test")))
    assert_round_trip(part_parser, Part(P.of_kind("test"), navigation=Navigation(1, 10, [EdgeTypes.delete])))
    assert_round_trip(part_parser, Part(P.of_kind("test"), "red", navigation=Navigation(1, 10, [EdgeTypes.delete])))
    with_clause = WithClause(WithClauseFilter("==", 0), Navigation(maybe_edge_types=[EdgeTypes.delete]))
    assert_round_trip(
        part_parser, Part(P.of_kind("test"), "green", with_clause, navigation=Navigation(1, 10, [EdgeTypes.delete]))
    )


def test_query() -> None:
    query = (
        Query.by(
            "ec2",
            P("cpu") > 4,
            (P("mem") < 23) | (P("mem") < 59) | P.context("foo[*]", P.single("mem") < 23, P.single("core") > 2),
        )
        .merge_with("cloud", Navigation(1, Navigation.Max, direction=Direction.inbound), Query.mk_term("cloud"))
        .traverse_out()
        .filter(P("some.int.value") < 1, P("some.other") == 23)
        .traverse_out()
        .filter(P("active") == 12, P.function("in_subnet").on("ip", "1.2.3.4/96"))
        .filter_with(WithClause(WithClauseFilter("==", 0), Navigation()))
        .group_by([AggregateVariable(AggregateVariableName("foo"))], [AggregateFunction("sum", "cpu")])
        .add_sort(Sort("test", "asc"))
        .with_limit(10)
    )
    assert str(query) == (
        'aggregate(foo: sum(cpu)):((is("ec2") and cpu > 4) and '
        "((mem < 23 or mem < 59) or foo[*].{(mem < 23 and core > 2)}))"
        ' {cloud: all <-default[1:]- is("cloud")} -default-> '
        "(some.int.value < 1 and some.other == 23) -default-> "
        '(active == 12 and in_subnet(ip, "1.2.3.4/96")) '
        "with(empty, -default->) sort test asc limit 10"
    )
    assert_round_trip(query_parser, query)


def test_special_queries() -> None:
    # unquoted date like test_date < @YESTERDAY
    assert str(parse_query("test_date < 2021-12-09")) == 'test_date < "2021-12-09"'


def test_query_with_preamble() -> None:
    parse_query('id("root")')  # no preamble
    # edge type can be defined in preamble
    q1 = parse_query('(edge_type=delete): id("root") -[0:1]->')
    assert q1.parts[0].navigation.edge_types == ["delete"]  # type: ignore
    # edge type can be defined via kwargs
    q2 = parse_query('id("root") -[0:1]->', edge_type="delete")
    assert q2.parts[0].navigation.edge_types == ["delete"]  # type: ignore
    # aggregation and preamble
    q3 = parse_query('aggregate(region: sum(cpu))(edge_type=delete): id("root") -[0:1]->')
    assert q3.aggregate.group_by[0].name == AggregateVariableName("region")  # type: ignore
    assert q3.aggregate.group_func[0].name == "cpu"  # type: ignore


def test_preamble_tags() -> None:
    assert preamble_tags_parser.parse("(edge_type=foo)") == {"edge_type": "foo"}
    assert preamble_tags_parser.parse("(edge_type=23)") == {"edge_type": 23}
    assert preamble_tags_parser.parse("(edge_type=23.123)") == {"edge_type": 23.123}
    assert preamble_tags_parser.parse("(edge_type=true)") == {"edge_type": True}
    assert preamble_tags_parser.parse("(edge_type=false)") == {"edge_type": False}
    assert preamble_tags_parser.parse('(edge_type="!@#*(&$({{:::")') == {"edge_type": "!@#*(&$({{:::"}
    assert preamble_parser.parse("") == (None, {})


def test_aggregate_group_variable() -> None:
    foo_var = AggregateVariableName("foo")
    bla_var = AggregateVariableName("bla")

    foo = aggregate_group_variable_parser.parse("foo")
    assert foo.name == foo_var
    assert foo.as_name is None

    bla = aggregate_group_variable_parser.parse("bla as bar")
    assert bla.name == bla_var
    assert bla.as_name == "bar"

    combined = aggregate_group_variable_parser.parse('"some_{foo}_{bla}_test" as foo_bar')
    assert combined.name == AggregateVariableCombined(["some_", foo_var, "_", bla_var, "_test"])
    assert combined.as_name == "foo_bar"


def test_aggregate_group_function() -> None:
    foo = aggregate_group_function_parser.parse("sum(foo)")
    assert foo.function == "sum"
    assert foo.name == "foo"
    assert foo.as_name is None
    bla = aggregate_group_function_parser.parse("sum(bla) as bar")
    assert foo.function == "sum"
    assert bla.name == "bla"
    assert bla.as_name == "bar"
    boo = aggregate_group_function_parser.parse("sum(boo * 1024.12 + 1) as bar")
    assert boo.ops == [("*", 1024.12), ("+", 1)]
    with pytest.raises(Exception):
        assert aggregate_group_function_parser.parse("sum(test / 3 +)")


def test_aggregate() -> None:
    agg: Aggregate = aggregate_parser.parse("aggregate(region: sum(cpu) as cpus, count(id) as instances)")
    assert len(agg.group_by) == 1
    assert agg.group_by[0].name == AggregateVariableName("region")
    assert len(agg.group_func) == 2
    assert agg.group_func[0].function == "sum"
    assert agg.group_func[0].name == "cpu"
    assert agg.group_func[0].as_name == "cpus"
    assert agg.group_func[1].function == "count"
    assert agg.group_func[1].name == "id"
    assert agg.group_func[1].as_name == "instances"
    agg2: Aggregate = aggregate_parser.parse("aggregate(count(id) as length)")
    assert agg2.group_by == []
    assert len(agg2.group_func) == 1


def test_sort_order() -> None:
    assert sort_parser.parse("sort foo") == [Sort("foo", "asc")]
    assert sort_parser.parse("sort foo asc") == [Sort("foo", "asc")]
    parsed = sort_parser.parse("sort foo asc, bla desc, bar")
    assert parsed == [Sort("foo", "asc"), Sort("bla", "desc"), Sort("bar", "asc")]
    assert_round_trip(query_parser, Query.by("test").add_sort(Sort("test")).add_sort(Sort("goo")))


def test_limit() -> None:
    assert limit_parser.parse("limit 23") == Limit(0, 23)
    assert limit_parser.parse("limit 3, 23") == Limit(3, 23)
    assert_round_trip(query_parser, Query.by("test").with_limit(23))


def test_with_clause() -> None:
    predicate_term.parse("foo == bla")
    wc: WithClause = with_clause_parser.parse("with(empty, -delete-> foo == bla and test > 23 with(any, -delete->))")
    assert wc.with_filter == WithClauseFilter("==", 0)
    assert wc.navigation == Navigation(maybe_edge_types=["delete"])
    assert str(wc.term) == '(foo == "bla" and test > 23)'
    assert str(wc.with_clause) == "with(any, -delete->)"
    term = Query.mk_term("foo", P("test") == 23)
    clause_filter = WithClauseFilter(">", 23)
    nav = Navigation()

    def edge(wc: WithClause) -> WithClause:
        wcr = evolve(wc, with_clause=edge(wc.with_clause)) if wc.with_clause else wc
        return evolve(wcr, navigation=evolve(wcr.navigation, maybe_edge_types=[EdgeTypes.default]))

    assert_round_trip(with_clause_parser, WithClause(clause_filter, nav, term, WithClause(clause_filter, nav)), edge)
    assert_round_trip(with_clause_parser, WithClause(clause_filter, nav), edge)


def test_special_cases() -> None:
    with pytest.raises(error.ParseError):
        # parser was able to read: is(instance) and sort in "stance_cores"
        parse_query("is(instance) and sort instance_cores")

    # parser read the reversed option as separate part, so following query became 3 parts
    q = parse_query("all sort kind desc limit 1 reversed -default-> all sort kind asc")
    assert len(q.parts) == 2


def test_usage_parser() -> None:
    def test_usage(s: str, start: datetime, end: Optional[datetime], metrics: List[str]) -> None:
        usage: WithUsage = with_usage_parser.parse(s)
        assert usage.start_from_now().timestamp() == approx(start.timestamp(), abs=1)
        if end:
            assert usage.end is not None
            assert usage.end_from_now().timestamp() == approx(end.timestamp(), abs=1)
        assert usage.metrics == metrics

    oneweekago = utc() - timedelta(weeks=1)
    twoweekago = utc() - timedelta(weeks=2)
    at = parse_utc("2023-06-10T12:23:21Z")
    test_usage("with_usage(1w, a,b,c)", oneweekago, None, ["a", "b", "c"])
    test_usage("with_usage(2w::1w, a,b,c)", twoweekago, oneweekago, ["a", "b", "c"])
    test_usage("with_usage(2w::1w, [a,b,c])", twoweekago, oneweekago, ["a", "b", "c"])
    test_usage("with_usage(2023-06-10T12:23:21Z::1w, [a,b,c])", at, oneweekago, ["a", "b", "c"])
    test_usage("with_usage(2023-06-10T12:23:21Z::2023-06-10T12:23:21Z, [a,b,c])", at, at, ["a", "b", "c"])


@given(query)
@settings(max_examples=200, suppress_health_check=list(HealthCheck))
def test_generated_query(q: Query) -> None:
    assert str(q) == str(parse_query(str(q)))


def assert_round_trip(parser: Parser, obj: object, after_parsed: Optional[Callable[[Any], Any]] = None) -> None:
    str_rep = str(obj)
    parsed = parser.parse(str_rep)
    post = after_parsed(parsed) if after_parsed else parsed
    assert str(post) == str_rep, f"Expected: {str(post)} but got {str_rep}.\nDifference: {DeepDiff(obj, post)}"
